@@ -6,7 +6,10 @@ use rlp::{Decodable, DecoderError, Encodable, Rlp, RlpStream};
 use sha3::{Digest, Keccak256};
 
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
-#[cfg_attr(feature = "with-codec", derive(codec::Encode, codec::Decode))]
+#[cfg_attr(
+	feature = "with-codec",
+	derive(codec::Encode, codec::Decode, scale_info::TypeInfo)
+)]
 #[cfg_attr(feature = "with-serde", derive(serde::Serialize, serde::Deserialize))]
 pub enum TransactionAction {
 	Call(H160),
@@ -39,7 +42,10 @@ impl Decodable for TransactionAction {
 }
 
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
-#[cfg_attr(feature = "with-codec", derive(codec::Encode, codec::Decode))]
+#[cfg_attr(
+	feature = "with-codec",
+	derive(codec::Encode, codec::Decode, scale_info::TypeInfo)
+)]
 #[cfg_attr(feature = "with-serde", derive(serde::Serialize, serde::Deserialize))]
 pub struct TransactionRecoveryId(pub u64);
 
@@ -70,6 +76,7 @@ impl TransactionRecoveryId {
 }
 
 #[derive(Clone, Debug, PartialEq, Eq)]
+#[cfg_attr(feature = "with-codec", derive(scale_info::TypeInfo))]
 #[cfg_attr(feature = "with-serde", derive(serde::Serialize, serde::Deserialize))]
 pub struct TransactionSignature {
 	v: TransactionRecoveryId,
@@ -161,7 +168,10 @@ impl codec::Decode for TransactionSignature {
 }
 
 #[derive(Clone, Debug, PartialEq, Eq)]
-#[cfg_attr(feature = "with-codec", derive(codec::Encode, codec::Decode))]
+#[cfg_attr(
+	feature = "with-codec",
+	derive(codec::Encode, codec::Decode, scale_info::TypeInfo)
+)]
 #[cfg_attr(feature = "with-serde", derive(serde::Serialize, serde::Deserialize))]
 pub struct AccessListItem {
 	pub address: Address,
@@ -188,7 +198,10 @@ impl Decodable for AccessListItem {
 pub type AccessList = Vec<AccessListItem>;
 
 #[derive(Clone, Debug, PartialEq, Eq)]
-#[cfg_attr(feature = "with-codec", derive(codec::Encode, codec::Decode))]
+#[cfg_attr(
+	feature = "with-codec",
+	derive(codec::Encode, codec::Decode, scale_info::TypeInfo)
+)]
 pub struct LegacyTransactionMessage {
 	pub nonce: U256,
 	pub gas_price: U256,
@@ -287,7 +300,11 @@ impl Encodable for EIP2930TransactionMessage {
 
 impl EIP2930TransactionMessage {
 	pub fn hash(&self) -> H256 {
-		H256::from_slice(Keccak256::digest(&rlp::encode(self)).as_slice())
+		let encoded = rlp::encode(self);
+		let mut out = alloc::vec![0; 1 + encoded.len()];
+		out[0] = 1;
+		out[1..].copy_from_slice(&encoded);
+		H256::from_slice(Keccak256::digest(&out).as_slice())
 	}
 }
 
@@ -322,7 +339,7 @@ impl From<EIP1559Transaction> for EIP1559TransactionMessage {
 
 impl Encodable for EIP1559TransactionMessage {
 	fn rlp_append(&self, s: &mut RlpStream) {
-		s.begin_list(8);
+		s.begin_list(9);
 		s.append(&self.chain_id);
 		s.append(&self.nonce);
 		s.append(&self.max_priority_fee_per_gas);
@@ -337,12 +354,19 @@ impl Encodable for EIP1559TransactionMessage {
 
 impl EIP1559TransactionMessage {
 	pub fn hash(&self) -> H256 {
-		H256::from_slice(Keccak256::digest(&rlp::encode(self)).as_slice())
+		let encoded = rlp::encode(self);
+		let mut out = alloc::vec![0; 1 + encoded.len()];
+		out[0] = 2;
+		out[1..].copy_from_slice(&encoded);
+		H256::from_slice(Keccak256::digest(&out).as_slice())
 	}
 }
 
 #[derive(Clone, Debug, PartialEq, Eq)]
-#[cfg_attr(feature = "with-codec", derive(codec::Encode, codec::Decode))]
+#[cfg_attr(
+	feature = "with-codec",
+	derive(codec::Encode, codec::Decode, scale_info::TypeInfo)
+)]
 #[cfg_attr(feature = "with-serde", derive(serde::Serialize, serde::Deserialize))]
 pub struct LegacyTransaction {
 	pub nonce: U256,
@@ -354,6 +378,12 @@ pub struct LegacyTransaction {
 	pub signature: TransactionSignature,
 }
 
+impl LegacyTransaction {
+	pub fn hash(&self) -> H256 {
+		H256::from_slice(Keccak256::digest(&rlp::encode(self)).as_slice())
+	}
+}
+
 impl Encodable for LegacyTransaction {
 	fn rlp_append(&self, s: &mut RlpStream) {
 		s.begin_list(9);
@@ -402,7 +432,10 @@ impl Decodable for LegacyTransaction {
 }
 
 #[derive(Clone, Debug, PartialEq, Eq)]
-#[cfg_attr(feature = "with-codec", derive(codec::Encode, codec::Decode))]
+#[cfg_attr(
+	feature = "with-codec",
+	derive(codec::Encode, codec::Decode, scale_info::TypeInfo)
+)]
 #[cfg_attr(feature = "with-serde", derive(serde::Serialize, serde::Deserialize))]
 pub struct EIP2930Transaction {
 	pub chain_id: u64,
@@ -418,6 +451,16 @@ pub struct EIP2930Transaction {
 	pub s: H256,
 }
 
+impl EIP2930Transaction {
+	pub fn hash(&self) -> H256 {
+		let encoded = rlp::encode(self);
+		let mut out = alloc::vec![0; 1 + encoded.len()];
+		out[0] = 1;
+		out[1..].copy_from_slice(&encoded);
+		H256::from_slice(Keccak256::digest(&out).as_slice())
+	}
+}
+
 impl Encodable for EIP2930Transaction {
 	fn rlp_append(&self, s: &mut RlpStream) {
 		s.begin_list(11);
@@ -466,7 +509,10 @@ impl Decodable for EIP2930Transaction {
 }
 
 #[derive(Clone, Debug, PartialEq, Eq)]
-#[cfg_attr(feature = "with-codec", derive(codec::Encode, codec::Decode))]
+#[cfg_attr(
+	feature = "with-codec",
+	derive(codec::Encode, codec::Decode, scale_info::TypeInfo)
+)]
 #[cfg_attr(feature = "with-serde", derive(serde::Serialize, serde::Deserialize))]
 pub struct EIP1559Transaction {
 	pub chain_id: u64,
@@ -483,6 +529,16 @@ pub struct EIP1559Transaction {
 	pub s: H256,
 }
 
+impl EIP1559Transaction {
+	pub fn hash(&self) -> H256 {
+		let encoded = rlp::encode(self);
+		let mut out = alloc::vec![0; 1 + encoded.len()];
+		out[0] = 2;
+		out[1..].copy_from_slice(&encoded);
+		H256::from_slice(Keccak256::digest(&out).as_slice())
+	}
+}
+
 impl Encodable for EIP1559Transaction {
 	fn rlp_append(&self, s: &mut RlpStream) {
 		s.begin_list(12);
@@ -544,6 +600,15 @@ pub enum TransactionV1 {
 	EIP2930(EIP2930Transaction),
 }
 
+impl TransactionV1 {
+	pub fn hash(&self) -> H256 {
+		match self {
+			TransactionV1::Legacy(t) => t.hash(),
+			TransactionV1::EIP2930(t) => t.hash(),
+		}
+	}
+}
+
 impl Encodable for TransactionV1 {
 	fn rlp_append(&self, s: &mut RlpStream) {
 		match self {
@@ -574,7 +639,10 @@ impl Decodable for TransactionV1 {
 }
 
 #[derive(Clone, Debug, PartialEq, Eq)]
-#[cfg_attr(feature = "with-codec", derive(codec::Encode, codec::Decode))]
+#[cfg_attr(
+	feature = "with-codec",
+	derive(codec::Encode, codec::Decode, scale_info::TypeInfo)
+)]
 #[cfg_attr(feature = "with-serde", derive(serde::Serialize, serde::Deserialize))]
 pub enum TransactionV2 {
 	/// Legacy transaction type
@@ -585,6 +653,16 @@ pub enum TransactionV2 {
 	EIP1559(EIP1559Transaction),
 }
 
+impl TransactionV2 {
+	pub fn hash(&self) -> H256 {
+		match self {
+			TransactionV2::Legacy(t) => t.hash(),
+			TransactionV2::EIP2930(t) => t.hash(),
+			TransactionV2::EIP1559(t) => t.hash(),
+		}
+	}
+}
+
 impl Encodable for TransactionV2 {
 	fn rlp_append(&self, s: &mut RlpStream) {
 		match self {
@@ -627,6 +705,27 @@ fn enveloped<T: Encodable>(id: u8, v: &T, s: &mut RlpStream) {
 	out.rlp_append(s)
 }
 
+impl From<LegacyTransaction> for TransactionV1 {
+	fn from(t: LegacyTransaction) -> Self {
+		TransactionV1::Legacy(t)
+	}
+}
+
+impl From<LegacyTransaction> for TransactionV2 {
+	fn from(t: LegacyTransaction) -> Self {
+		TransactionV2::Legacy(t)
+	}
+}
+
+impl From<TransactionV1> for TransactionV2 {
+	fn from(t: TransactionV1) -> Self {
+		match t {
+			TransactionV1::Legacy(t) => TransactionV2::Legacy(t),
+			TransactionV1::EIP2930(t) => TransactionV2::EIP2930(t),
+		}
+	}
+}
+
 #[cfg(test)]
 mod tests {
 	use super::*;
